/**
  *  \file client/dialogs/buildqueue.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/dialogs/buildqueue.h"
#include "client/actions.h"
#include "client/dialogs/buildqueuesummary.h"
#include "client/screens.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/smallset.h"
#include "game/actions/changebuildqueue.h"
#include "game/global.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "int/if/playerif.h"
#include "int/if/userif.h"
#include "int/value.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/quit.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "util/unicodechars.h"

namespace {
    /*
     *  BuildQueueList - a list box displaying the build queue
     */

    const int ICON_HEMS = 3;       // half ems
    const int ACTION_EMS = 25;
    const int FCODE_EMS = 5;
    const int QPOS_EMS = 5;
    const int POINTS_EMS = 10;
    const int GAP_PX = 5;
    const int PAD_PX = 2;

    class BuildQueueList : public UIListbox {
     public:
        typedef GChangeBuildQueue::Infos_t Infos_t;

        enum Column {
            QueuePositionColumn,
            BuildPointsColumn
        };
        typedef SmallSet<Column> Columns_t;

        BuildQueueList(Columns_t cols);

        void setContent(const Infos_t& data);
        void scrollToPlanet(int planetId);

        const Infos_t& getContent() const;
        bool hasChanges() const;
        int getCurrentPlanetId() const;

        virtual void drawContent(GfxCanvas& can);
        virtual void drawPart(GfxCanvas& can, int from, int to);

     private:
        Infos_t m_data;
        Columns_t m_columns;
    };

    class BuildQueueKeyHandler : public UIFixedWidget {
     public:
        BuildQueueKeyHandler(GChangeBuildQueue& action, BuildQueueList& list);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void drawContent(GfxCanvas& can);

     private:
        void update();

        GChangeBuildQueue& m_action;
        BuildQueueList& m_list;
    };

    class BuildQueueDialog {
     public:
        typedef GChangeBuildQueue::Infos_t Infos_t;

        BuildQueueDialog(/*ui::Root& root, afl::string::Translator& tx,*/ GChangeBuildQueue& action, BuildQueueList::Columns_t cols)
            : m_list(cols),
              m_action(action),
              m_targetScreen(),
              m_targetId()
            { }

        void setContent(const Infos_t& data)
            { m_list.setContent(data); }

        void scrollToPlanet(int planetId)
            { m_list.scrollToPlanet(planetId); }

        void run()
            {
                ObjectHolder del;
                UIWindow win(_("Manage Build Queue"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
                win.setLayout(UIVBoxLayout::instance5);
                win.add(UIFrameGroup::wrap(del, m_list));

                UIWidget& keys = del.add(new BuildQueueKeyHandler(m_action, m_list));
                win.add(keys);

                UIGroup& g = del.add(new UIGroup(UIHBoxLayout::instance5));
                UIButton& btnFaster = del.add(new UIButton("+", '+'));
                g.add(btnFaster);
                g.add(del.add(new UIStaticText(_("Build earlier"), UIStaticText::StaticNormal)));

                UIButton& btnSlower = del.add(new UIButton("-", '-'));
                g.add(btnSlower);
                g.add(del.add(new UIStaticText(_("Build later"), UIStaticText::StaticNormal)));
                g.add(del.add(new UISpacer()));

                UIButton& btnGoto = del.add(new UIButton(0, _("Go to"), 'g'));
                g.add(btnGoto);
                btnGoto.sig_fired.add(this, &BuildQueueDialog::onGoto);

                UIButton& btnSummary = del.add(new UIButton(0, _("Summary..."), 's'));
                g.add(btnSummary);
                btnSummary.sig_fired.add(this, &BuildQueueDialog::onSummary);

                win.add(g);

                UIButton& btnOK = del.add(new UIButton(1, _("OK"), SDLK_RETURN));
                UIButton& btnCancel = del.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE));
                UIGroup& btns = del.add(new UIGroup(UIHBoxLayout::instance5));
                btns.add(btnOK);
                btns.add(btnCancel);
                btns.add(del.add(new UISpacer()));
                btns.add(del.add(new UIButton(_("Help"), 'h')));
                win.add(btns);
                btnOK.addStop();
                btnCancel.addStop();

                win.add(del.add(new UIQuit(0)));
                win.add(del.add(new WHelpWidget("pcc2:queuemanager")));

                win.pack();
                ui_root->centerWidget(win);
                if (ui_root->runModal(win)) {
                    m_action.commit();
                }
            }

        void onSummary()
            {
                doBuildQueueSummaryDialog(m_list.getContent());
            }

        int getTargetScreen() const
            { return m_targetScreen; }

        int getTargetId() const
            { return m_targetId; }

        void onGoto()
            {
                // Fail-safe
                int id = m_list.getCurrentPlanetId();
                if (id == 0) {
                    return;
                }

                // Ask for confirmation
                enum { Yes, No, Cancel };
                int mode;
                if (m_list.hasChanges()) {
                    UIKeyString yes(_("Yes")), no(_("No"));
                    mode = UIMessageBox(_("Apply changes?"), _("Manage Build Queue"))
                        .addButton(Yes,    yes.getString(), yes.getKey())
                        .addButton(No,     no.getString(),  no.getKey())
                        .addButton(Cancel, _("Cancel"), SDLK_ESCAPE)
                        .addKey(Yes, SDLK_RETURN)
                        .addKey(Yes, ' ')
                        .run();
                } else {
                    mode = Yes;
                }

                // Do it
                if (mode != Cancel) {
                    m_targetScreen = 3;
                    m_targetId = id;
                    ui_root->stop(mode == Yes);
                }
            }

     private:
        BuildQueueList m_list;
        GChangeBuildQueue& m_action;
        int m_targetScreen;
        int m_targetId;
    };

    int getWidth(BuildQueueList::Columns_t cols)
    {
        int em = font_ems[FONT_NORMAL];
        int extraSize = 0;
        if (cols.contains(BuildQueueList::QueuePositionColumn)) {
            extraSize += em * QPOS_EMS + GAP_PX;
        }
        if (cols.contains(BuildQueueList::BuildPointsColumn)) {
            extraSize += em * POINTS_EMS + GAP_PX;
        }

        return (em * (ACTION_EMS + FCODE_EMS) + 3*GAP_PX + extraSize);
    }
}

/*
 *  BuildQueueList
 */

BuildQueueList::BuildQueueList(Columns_t cols)
    : UIListbox(0, 10, getWidth(cols), 0),
      m_data(),
      m_columns(cols)
{
    setHeaderHeight(font_heights[FONT_NORMAL]);
    setItemHeight(font_heights[FONT_NORMAL] + font_heights[FONT_SMALL] + 2*PAD_PX);
}

void
BuildQueueList::setContent(const Infos_t& data)
{
    // Remember current Id
    int currentId = getCurrentPlanetId();

    // Update
    m_data = data;
    drawWidget();
    setSize(m_data.size());
    adjustPosition();

    // Select current Id
    if (currentId != 0) {
        scrollToPlanet(currentId);
    }
}

void
BuildQueueList::scrollToPlanet(int planetId)
{
    for (size_t i = 0, n = m_data.size(); i < n; ++i) {
        if (m_data[i].planetId == planetId) {
            scrollTo(int(i), 1);
            break;
        }
    }
}

const BuildQueueList::Infos_t&
BuildQueueList::getContent() const
{
    return m_data;
}

bool
BuildQueueList::hasChanges() const
{
    for (size_t i = 0, n = m_data.size(); i < n; ++i) {
        if (m_data[i].isChange) {
            return true;
        }
    }
    return false;
}

int
BuildQueueList::getCurrentPlanetId() const
{
    int currentId = 0;
    int currentItem = getCurrent();
    if (currentItem < int(m_data.size())) {
        currentId = m_data[currentItem].planetId;
    }
    return currentId;
}

void
BuildQueueList::drawContent(GfxCanvas& can)
{
    const int em = font_ems[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setColor(UIColor::tc_Static);
    ctx.useFont(*fonts[FONT_NORMAL]);

    GfxRect extent = getExtent();
    GfxRect area = extent.splitY(getHeaderHeight());

    drawHLine(ctx, area.x, area.getBottomY()-1, area.getRightX()-1);

    area.consumeX(GAP_PX + ICON_HEMS*em/2);
    outTextF(ctx, area.splitX(ACTION_EMS * em + GAP_PX), _("Build Order"));
    outTextF(ctx, area.splitX(FCODE_EMS * em + GAP_PX), _("FCode"));
    if (m_columns.contains(QueuePositionColumn)) {
        outTextF(ctx, area.splitX(QPOS_EMS * em + GAP_PX), _("Q-Pos"));
    }
    if (m_columns.contains(BuildPointsColumn)) {
        outTextF(ctx, area, _("Build Points"));
    }
    UIListbox::drawContent(can);
}

void
BuildQueueList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        size_t item = i + getTop();
        if (item < m_data.size()) {
            const GChangeBuildQueue::Info& e = m_data[item];
            const int em = font_ems[FONT_NORMAL];

            area.consumeY(PAD_PX);
            ctx.useFont(*fonts[FONT_NORMAL]);

            // Icon
            GfxRect iconArea = area.splitX(GAP_PX + ICON_HEMS*em/2);
            ctx.setTextAlign(1, 0);
            if (e.planned) {
                ctx.setColor(UIColor::tc_Faded);
                outTextF(ctx, iconArea, UTF_STOPWATCH);
            } else {
                ctx.setColor(UIColor::tc_Green);
                outTextF(ctx, iconArea, UTF_CHECK_MARK);
            }

            // Name
            UIColor::Color defColor = e.planned ? UIColor::tc_Faded : UIColor::tc_Static;
            GfxRect nameArea = area.splitX(ACTION_EMS * em);
            ctx.setTextAlign(0, 0);
            ctx.setColor(defColor);
            outTextF(ctx, nameArea.splitY(font_heights[FONT_NORMAL]), e.actionName);
            ctx.useFont(*fonts[FONT_SMALL]);
            ctx.setColor(UIColor::tc_Faded);
            outTextF(ctx, nameArea, format("(%s, #%d)", e.planetName, e.planetId));
            area.consumeX(GAP_PX);

            // Friendly code
            ctx.useFont(e.hasPriority ? *fonts[FONT_NORMAL_BOLD] : *fonts[FONT_NORMAL]);
            ctx.setColor(e.conflict ? UIColor::tc_Red : defColor);
            outTextF(ctx, area.splitX(FCODE_EMS * em), e.friendlyCode);
            area.consumeX(GAP_PX);

            // Queue position
            if (m_columns.contains(QueuePositionColumn)) {
                GfxRect queueArea = area.splitX(QPOS_EMS * em);
                if (e.queuePosition != 0) {
                    ctx.useFont(*fonts[FONT_NORMAL]);
                    ctx.setColor(defColor);
                    ctx.setTextAlign(2, 0);
                    outTextF(ctx, queueArea, format("%d", e.queuePosition));
                }
                area.consumeX(GAP_PX);
            }

            // Points
            if (m_columns.contains(BuildPointsColumn)) {
                int space = POINTS_EMS * em;
                int half = space/2 - 3;
                GfxRect needArea = area.splitX(half);
                GfxRect haveArea = area.splitX(space - half);

                if (e.pointsRequired.isKnown()) {
                    int32_t reqd = e.pointsRequired;
                    ctx.useFont(*fonts[FONT_NORMAL]);
                    if (e.pointsAvailable.isKnown()) {
                        int32_t avail = e.pointsAvailable;
                        ctx.setColor(defColor);
                        ctx.setTextAlign(0, 0);
                        outTextF(ctx, haveArea, format(" / %d", avail));

                        ctx.setColor(e.hasPriority ? reqd > avail ? UIColor::tc_Red : defColor : UIColor::tc_Faded);
                    } else {
                        ctx.setColor(defColor);
                    }
                    ctx.setTextAlign(2, 0);
                    outTextF(ctx, needArea, format("%d", reqd));
                }
            }
        }
    }
}


/*
 *  BuildQueueKeyHandler - special keys for the dialog
 */

BuildQueueKeyHandler::BuildQueueKeyHandler(GChangeBuildQueue& action, BuildQueueList& list)
    : m_action(action),
      m_list(list)
{ }

bool
BuildQueueKeyHandler::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        uint32_t key = event.key.code;
        if (key == '-' || key == SDLK_DOWN + ss_Shift) {
            m_action.decreasePriority(m_list.getCurrent());
            update();
            return true;
        } else if (key == '+' || key == SDLK_UP + ss_Shift) {
            m_action.increasePriority(m_list.getCurrent());
            update();
            return true;
        } else if (key >= '0' && key <= '9') {
            m_action.setPriority(m_list.getCurrent(), key - '0');
            update();
            return true;
        }
    }
    return UIFixedWidget::handleEvent(event, second_pass);
}

void
BuildQueueKeyHandler::drawContent(GfxCanvas& /*can*/)
{ }

void
BuildQueueKeyHandler::update()
{
    GChangeBuildQueue::Infos_t infos;
    m_action.describe(infos);
    m_list.setContent(infos);
}



/*
 *  Main entry point
 */


void doBuildQueueDialog(GGameTurn& turn, int baseId)
{
    // Simplest way to get my PBPs (also used in specbuildinfo):
    std::auto_ptr<IntValue> pbpValue(getPlayerProperty(getPlayerId(), iplPBPs));
    mn32_t pbps;
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(pbpValue.get())) {
        pbps = mn32_t(iv->getValue());
    }

    // Set up action
    GChangeBuildQueue action(turn, host, config, myRandom, getPlayerId());
    GChangeBuildQueue::Infos_t infos;
    action.addPlannedBuilds(getProcessList());
    action.setAvailableBuildPoints(pbps);
    action.describe(infos);
    if (infos.empty()) {
        messageBox(_("You have no active ship build orders."),
                   _("Manage Build Queue"));
        return;
    }

    // Column configuration
    BuildQueueList::Columns_t cols;
    for (size_t i = 0, n = infos.size(); i < n; ++i) {
        if (infos[i].queuePosition != 0) {
            cols += BuildQueueList::QueuePositionColumn;
        }
        if (infos[i].pointsRequired.isKnown()) {
            cols += BuildQueueList::BuildPointsColumn;
        }
    }

    // Set up dialog
    BuildQueueDialog dlg(action, cols);
    dlg.setContent(infos);
    dlg.scrollToPlanet(baseId);
    dlg.run();

    if (int screen = dlg.getTargetScreen()) {
        postGoToScreen(IntUserInterfaceBinding::get(), screen, dlg.getTargetId());
    }
}
