/**
  *  \file client/dialogs/helpdlg.cc
  *  \brief Help Viewer
  *
  *  The PCC2 Help Viewer displays XML rendered into a RichDocument
  *  (as opposed to PCC 1.x, which displays preformatted text with
  *  custom control codes from a .res file).
  *
  *  This module provides user interface and control.
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/helpdlg.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "io/streams.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/richdocument.h"
#include "ui/richhelp.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/xmlread.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/quit.h"
#include "util/helpindex.h"
#include "ui/richparser.h"

namespace {
    enum {
        id_Back = 100,
        id_Toc = 101
    };

    const char* TOC_PAGE = "toc";

    /** Help dialog. */
    class WHelpDialog : public UIWindow {
        ObjectHolder h;
        UIRichDocument doc;
        string_t currentPage;

        /* History: this is a ring buffer. historyCount contains the number
           of valid elements, historyHead points to the next usable (that is,
           unless the buffer is full, an unused) element. We limit history by
           simply not increasing historyCount above MAX. */
        struct History {
            string_t page;
            int top;
            RichDocument::linkid_t link;
        };
        enum { MAX = 32 };
        History history[MAX];
        int historyCount;
        int historyHead;

        void init();
        void setBackState();
        void pushHistory();
     public:
        WHelpDialog();
        void loadPage(string_t page);
        void onLinkClick(string_t page);
        void onBack();
        void onGoTo();
        bool handleEvent(const UIEvent& event, bool second_pass);
    };

    RichText richReplace(const RichText& outer, const RichText& inner)
    {
        string_t::size_type pos = outer.getText().find("%s");
        if (pos != string_t::npos) {
            return outer.substr(0, pos) + inner + outer.substr(pos+2);
        } else {
            return outer;
        }
    }
}

/** Constructor. */
WHelpDialog::WHelpDialog()
    : UIWindow(_("Help"), UIWindow::GREEN_WINDOW, COLOR_GRAY),
      doc(font_sizes[FONT_NORMAL].scaledBy(40, 20), UIRichDocument::fl_Help),
      currentPage(),
      historyCount(0),
      historyHead(0)
{
    init();
}

/** Build the dialog. */
void
WHelpDialog::init()
{
    // VBox
    //   HBox
    //     UIRichDocument
    //     UIScrollbar
    //   HBox
    //     UIButton...
    setLayout(UIVBoxLayout::instance5);

    /* We use instance5 for g1 to have a little room between the document and the scrollbar.
       Normally, we don't have room between scrollee and scrollbar, but here it serves as
       a margin. */
    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(doc);
    g1.add(h.add(new UIScrollbar(doc)));
    doc.sig_link_click.add(this, &WHelpDialog::onLinkClick);

    g2.add(h.add(new UIButton(id_Back, _("Back"), SDLK_BACKSPACE)).addKey());
    g2.add(h.add(new UIButton(id_Toc, _("T - Content"), 't')).addKey());
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());

    add(g1);
    add(g2);
    add(h.add(new UIQuit(1)));
    pack();

    setBackState();
}

/** Set state of "Back" button. It is only enabled when we actually
    have history. */
void
WHelpDialog::setBackState()
{
    setStateOf(id_Back, st_Disabled, historyCount == 0);
    setStateOf(id_Toc, st_Disabled, currentPage == TOC_PAGE);
}

/** Push current location on the history. */
void
WHelpDialog::pushHistory()
{
    // Anything to save?
    if (currentPage.empty())
        return;

    // Fill in entry
    History& h = history[historyHead];
    h.page = currentPage;
    h.top  = doc.getTop();
    h.link = doc.getSelectedLink();

    // Advance
    historyHead = (historyHead+1) % MAX;
    if (historyCount < MAX)
        ++historyCount;
}

/** Load a page. */
void
WHelpDialog::loadPage(string_t page)
{
    // Clear document
    RichDocument& d = doc.getDocument();
    d.clear();

    // Load all applicable pages
    std::vector<HelpIndex::Node*> nodes;
    HelpIndex::instance.find(page, nodes);
    if (nodes.empty()) {
        /* Render error page */
        d.add(RichText(_("Error")).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        d.addParagraph();
        d.add(richReplace(_("The requested page \"%s\" could not be found."), RichText(page).withAttribute(RichTextStyleAttribute::Bold)));
        d.addParagraph();
        d.add(RichParser::textFromXml(_("<b>See also:</b>&#160;<a href=\"toc\">Help Content</a>")));
    } else {
        /* Render individual pages */
        // FIXME: we could be more explicit about the origins
        for (size_t i = 0, n = nodes.size(); i < n; ++i) {
            const string_t fileName = nodes[i]->file.name;
            Ptr<Stream> file = Stream::openFileNT(fileName, Stream::C_READ);
            if (!file) {
                d.add(RichText(_("Error")).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
                d.addParagraph();
                d.add(richReplace(_("Help file \"%s\" could not be opened."), RichText(fileName).withAttribute(RichTextStyleAttribute::Bold)));
                d.addParagraph();
            } else {
                /* Found. Read and render it. */
                file->seek(nodes[i]->pos);
                XmlReader rdr(*file);
                rdr.read();                // This will produce a "<page"
                rdr.read();                // This will produce an attribute (or plaintext).
                loadHelpPage(d, rdr);
                d.addParagraph();
                if (n != 1 || nodes[i]->file.origin != "(PCC2)") {
                    d.addRight(d.getPageWidth(), RichText(UIColor::tc_Faded, format(_("from %s"), Directory::getFileName(fileName))).withAttribute(RichTextStyleAttribute::Small));
                    d.addParagraph();
                }
            }
        }
    }
    d.finish();

    // Update widget
    doc.handleDocumentUpdate();
    currentPage = page;
}

/** Handle click on a link. */
void
WHelpDialog::onLinkClick(string_t page)
{
    pushHistory();
    loadPage(page);
    setBackState();
}

/** Handle "Back" button. */
void
WHelpDialog::onBack()
{
    if (historyCount > 0) {
        // Pop entry
        --historyCount;
        historyHead = (historyHead > 0 ? historyHead-1 : MAX-1);
        const History& h = history[historyHead];

        // Load that page
        loadPage(h.page);
        doc.scrollTo(h.top, 1);
        if (doc.getDocument().isLinkVisible(h.link, GfxRect(0, doc.getTopY(), doc.getExtent().w, doc.getExtent().h)))
            doc.setSelectedLink(h.link);
        setBackState();
    }
}

void
WHelpDialog::onGoTo()
{
    string_t adr(currentPage);
    if (UIInputLine::run(adr, 4096, _("Go to"), _("Enter page name:"), 20, 0)) {
        onLinkClick(adr);
    }
}

bool
WHelpDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_BACKSPACE:
         case 'b':
            onBack();
            return true;
         case 'g':
            onGoTo();
            return true;
         case 'h':
         case ss_Alt + 'h':
         case SDLK_F1:
            onLinkClick("help");
            return true;
         case 't':
         case ss_Shift + SDLK_F1:
            onLinkClick(TOC_PAGE);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

/** Help, main entry point. Displays the help viewer, starting on the
    specified page. */
void
doHelp(string_t page)
{
    WHelpDialog dlg;
    dlg.loadPage(page);

    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}
