/**
  *  \file client/dlg-subject.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-subject.h"
#include "ui/layout.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "util/misc.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "client/widgets/msgfilter.h"
#include "ui/widgets/quit.h"

/*! \class WSubjectList
    \brief Message Subject List

    This widget displays a list of all message subjects of a
    GMessagebox. */

/** Constructor.
    \param id      Widget Id
    \param mbox    Mailbox box to display
    \param filter  Filter to apply to mailbox */
WSubjectList::WSubjectList(int id, GMessagebox& mbox, GMessageFilter& filter)
    : UIListbox(id, 20, 25 * font_ems[FONT_NORMAL], 0),
      mbox(mbox),
      filter(filter),
      groups()
{
    createGroupList();
}

WSubjectList::~WSubjectList()
{ }

/** Create the group list. */
void
WSubjectList::createGroupList()
{
    std::vector<Group> new_groups;

    int this_group_start = -1;
    string_t this_group_heading;

    for (int i = 0, max = mbox.getCount(); i < max; ++i) {
        string_t this_msg_heading = mbox.getHeading(i);
        if (this_msg_heading != this_group_heading) {
            if (this_group_start >= 0)
                new_groups.push_back(Group(this_group_start, i - this_group_start));
            this_group_heading = this_msg_heading;
            this_group_start = i;
        }
    }
    if (this_group_start >= 0)
        new_groups.push_back(Group(this_group_start, mbox.getCount() - this_group_start));

    new_groups.swap(groups);
    setSize(groups.size());
}


void
WSubjectList::toggleFilterSelectedMessage()
{
    if (size_t(getCurrent()) < groups.size()) {
        filter.toggleFilter(mbox.getHeading(groups[getCurrent()].first_msg));
        drawWidget();
    }
}

bool
WSubjectList::isSelectedMessageFiltered() const
{
    if (size_t(getCurrent()) < groups.size()) {
        return filter.isFiltered(mbox.getHeading(groups[getCurrent()].first_msg));
    } else {
        return false;
    }
}

/** Get selected message. Since each row can represent multiple
    messages, this returns the first message of the row. */
int
WSubjectList::getSelectedMessage() const
{
    if (size_t(getCurrent()) < groups.size()) {
        return groups[getCurrent()].first_msg;
    } else {
        return 0;
    }
}

/** Select a message. Selects the row which contains the message. */
void
WSubjectList::selectMessage(int n)
{
    for (std::vector<Group>::size_type i = 0; i < groups.size(); ++i) {
        if (groups[i].first_msg <= n && n - groups[i].first_msg < groups[i].count) {
            scrollTo(i, 1);
            break;
        }
    }
}

void
WSubjectList::regenerate()
{
    int n = getSelectedMessage();
    createGroupList();
    selectMessage(n);
    drawWidget();
    sig_moved.raise(getId());
}

void
WSubjectList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        const size_t index = size_t(getTop() + i);
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (index < groups.size()) {
            const int divi = 3*font_ems[FONT_NORMAL];
            if (filter.isFiltered(mbox.getHeading(groups[index].first_msg)))
                ctx.setColor(UIColor::tc_Faded);
            ctx.setTextAlign(2, 0);
            outTextF(ctx, area.splitX(divi), numToString(groups[index].count));
            ctx.setTextAlign(0, 0);
            area.consumeX(5);
            outTextF(ctx, area, mbox.getHeading(groups[index].first_msg));
        }
    }
}

/** Do a standard dialog.
    \return nonzero if user chose OK */
int
WSubjectList::doStandardDialog()
{
    ObjectHolder h;
    
    UIWindow window(_("Message Subjects"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    window.setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& group_list = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    group_list.add(*this);
    group_list.add(h.add(new UIScrollbar(*this)));

    group_buttons.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));

    window.add(group_list);
    window.add(h.add(new WMessageFilterControl(*this)));
    window.add(group_buttons);
    window.add(h.add(new UIQuit(0)));
    window.pack();

    setFocusedWidget();

    ui_root->centerWidget(window);
    ui_root->add(window);
    int result = ui_root->run();
    ui_root->close(window);

    return result;
}
