/**
  *  \file game/stat-file.cc
  *  \brief Score Statistics - Global Instance
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "game/gamedirs.h"
#include "game/stat.h"
#include "io/streams.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "game/gen.h"
#include "util/modinit.h"

static GStatFile* statistics_file;

/** Initialize score statistics module. Loads the statistics file if there is one.
    \post isStatisticsFileLoaded() */
static void
initStatisticsFile()
{
    if (statistics_file == 0) {
        statistics_file = new GStatFile();
        Ptr<Stream> s;

        try {
            if ((s = game_file_dir->openFileNT("score.cc", Stream::C_READ)) != 0) {
                /* we have a score.cc file, load it */
                statistics_file->load(*s);
            } else if ((s = game_file_dir->openFileNT("stat.cc", Stream::C_READ)) != 0) {
                /* we have a stat.cc file, load it */
                statistics_file->loadOldFile(*s);
            } else {
                /* no file, start anew */
                statistics_file->init();
            }
        }
        catch (FileProblemException& e) {
            console.write(LOG_ERROR, format(_("%s: %s, file has been ignored"),
                                            e.getFileName(),
                                            e.what()));
            statistics_file->init();
        }
    }
}

/** Save score statistics. */
static void
saveStatisticsFile()
{
    if (statistics_file != 0) {
        if (!statistics_file->canSaveSafely()) {
            console.write(LOG_WARN, _("The statistics file in game directory was written by a newer version of PCC2; changes not written."));
        } else {
            Ptr<Stream> s = game_file_dir->openFileNT("score.cc", Stream::C_CREATE);
            if (s != 0) {
                statistics_file->save(*s);
            } else {
                console.write(LOG_ERROR, _("Unable to create statistics file; score database not written."));
            }
        }
    }
}

/** Shut down score statistics module.
    \post !isStatisticsFileLoaded() */
static void
doneStatisticsFile()
{
    delete statistics_file;
    statistics_file = 0;
}

/** Check whether the statistics file has been loaded. */
bool
isStatisticsFileLoaded()
{
    return statistics_file != 0;
}

/** Get currently loaded statistics file.
    \pre isStatisticsFileLoaded() */
GStatFile&
getStatistics()
{
    ASSERT(isStatisticsFileLoaded());
    return *statistics_file;
}

extern const ModuleInitializer mod_stat[] = {
    { OnEnterGame, initStatisticsFile },
    { OnLeaveGame, saveStatisticsFile },
    { OnLeftGame, doneStatisticsFile },
    { 0, 0 },
};
