/**
  *  \file int/globalif.cc
  *  \brief Interpreter Game Interface: Global Functions and Properties
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include <cmath>
#include <ctime>
#include <limits.h>
#include "int/if/globalif.h"
#include "game/coord.h"
#include "int/data.h"
#include "int/value.h"
#include "int/compound.h"
#include "int/context.h"
#include "int/interror.h"
#include "int/if/ifutil.h"
#include "game/struct.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/config.h"
#include "game/team.h"
#include "version.h"
#include "game/host.h"
#include "cpluslib/misc.h"
#include "int/intglobal.h"
#include "game/fcode.h"
#include "int/if/playerif.h"
#include "game/cmds.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/reg.h"
#include "game/cspec.h"
#include "game/drawing.h"
#include "int/args.h"
#include "client/marks.h"
#include "util/random.h"
#include "game/sel.h"
#include "int/if/userif.h"
#include "int/contextio.h"
#include "game/gamedirs.h"
#include "game/storage.h"
#include "game/selio.h"
#include "util/answer.h"
#include "int/file.h"
#include "int/expr.h"
#include "int/exec.h"
#include "int/process.h"
#include "int/notify.h"
#include "int/propenum.h"
#include "game/actions/preconditions.h"
#include "game/gen.h"
#include "game/pref.h"

class TextFile;

namespace {
    enum { GlobalProperty, MyPlayerProperty, UIProperty };
}

static const TIntMapping global_mapping[] = {
    /* @q Chart.X:Int (Global Property), Chart.Y:Int (Global Property)
       Current position in the starchart.
       - on the starchart (and player screen): center of starchart.
         Values can be assigned to modify the current position.
       - on control screens: center of scanner (=location of current unit).
         Cannot be modified.
       @diff PCC 1.x always used the "on the starchart" interpretation, i.e. these values
       give a starchart position that may not have anything to do with the current unit's
       position the user actually sees.
       @assignable
       @see UI.GotoChart
       @since PCC2 1.99.10, PCC 1.0.16 */
    { "CHART.X",               iuiChartX,            UIProperty,       thInt },
    { "CHART.Y",               iuiChartY,            UIProperty,       thInt },
    { "MY.BASES",              iplScoreBases,        MyPlayerProperty, thInt },
    { "MY.INMSGS",             igpMyInMsgs,          GlobalProperty,   thInt },
    { "MY.OUTMSGS",            igpMyOutMsgs,         GlobalProperty,   thInt },
    { "MY.PBPS",               iplPBPs,              MyPlayerProperty, thInt },
    { "MY.PLANETS",            iplScorePlanets,      MyPlayerProperty, thInt },
    { "MY.RACE",               iplShortName,         MyPlayerProperty, thString },
    { "MY.RACE$",              iplId,                MyPlayerProperty, thInt },
    { "MY.RACE.ADJ",           iplAdjName,           MyPlayerProperty, thString },
    { "MY.RACE.FULL",          iplFullName,          MyPlayerProperty, thString },
    { "MY.RACE.ID",            iplRaceId,            MyPlayerProperty, thInt },
    { "MY.RACE.MISSION",       iplMission,           MyPlayerProperty, thInt },
    { "MY.SCORE",              iplScore,             MyPlayerProperty, thInt },
    { "MY.SHIPS",              iplScoreShips,        MyPlayerProperty, thInt },
    { "MY.SHIPS.CAPITAL",      iplScoreCapital,      MyPlayerProperty, thInt },
    { "MY.SHIPS.FREIGHTERS",   iplScoreFreighters,   MyPlayerProperty, thInt },
    { "MY.TEAM",               iplTeam,              MyPlayerProperty, thInt },
    { "MY.VCRS",               igpMyVCRs,            GlobalProperty,   thInt },
    { "SELECTION.LAYER",       igpSelectionLayer,    GlobalProperty,   thInt },
    { "SHIPS.CAPITAL",         iplTotalCapital,      MyPlayerProperty, thInt },  // Implemented as player properties because
    { "SHIPS.FREIGHTERS",      iplTotalFreighters,   MyPlayerProperty, thInt },  // that has easier access to a 'TGen'. A better
    { "SHIPS.TOTAL",           iplTotalShips,        MyPlayerProperty, thInt },  // way may be to implement it using GStatFile.
    // SYSTEM.EGG
    // SYSTEM.ERR -> global variable (localizable!)
    { "SYSTEM.GAMEDIRECTORY",  igpGameDirectory,     GlobalProperty,   thString },
    { "SYSTEM.GAMETYPE",       igpRegSharewareText,  GlobalProperty,   thString },
    { "SYSTEM.GAMETYPE$",      igpRegSharewareFlag,  GlobalProperty,   thInt },
    // SYSTEM.GUI -> conif, guiif
    { "SYSTEM.HOST",           igpSystemHost,        GlobalProperty,   thString },
    { "SYSTEM.HOST$",          igpSystemHostCode,    GlobalProperty,   thInt },
    { "SYSTEM.HOSTVERSION",    igpSystemHostVersion, GlobalProperty,   thString },
    { "SYSTEM.LANGUAGE",       igpSystemLanguage,    GlobalProperty,   thString },
    { "SYSTEM.LOCAL",          igpFileFormatLocal,   GlobalProperty,   thInt },
    { "SYSTEM.PROGRAM",        igpSystemProgram,     GlobalProperty,   thString },
    { "SYSTEM.RANDOMSEED",     igpRandomSeed,        GlobalProperty,   thInt },
    { "SYSTEM.REGSTR1",        igpRegStr1,           GlobalProperty,   thString },
    { "SYSTEM.REGSTR2",        igpRegStr2,           GlobalProperty,   thString },
    { "SYSTEM.REMOTE",         igpFileFormatRemote,  GlobalProperty,   thInt },
    { "SYSTEM.ROOTDIRECTORY",  igpRootDirectory,     GlobalProperty,   thString },
    /* @q System.Sim:Bool (Global Property)
       True if the combat simulator is currently in use, otherwise false.
       @since PCC2 1.99.10, PCC 1.0.9 */
    { "SYSTEM.SIM",            iuiSimFlag,           UIProperty,       thBool },
    { "SYSTEM.VERSION",        igpSystemVersion,     GlobalProperty,   thString },
    { "SYSTEM.VERSION$",       igpSystemVersionCode, GlobalProperty,   thInt },
    { "TURN",                  igpTurnNumber,        GlobalProperty,   thInt },
    { "TURN.DATE",             igpTurnDate,          GlobalProperty,   thString },
    { "TURN.ISNEW",            igpTurnIsNew,         GlobalProperty,   thBool },
    { "TURN.TIME",             igpTurnTime,          GlobalProperty,   thString },
    /* @q UI.Screen:Int (Global Property)
       Number of current screen. See {UI.GotoScreen} for a list.
       0 if no control screen is active.
       @see UI.GotoScreen
       @since PCC2 1.99.10, PCC 1.0.14 */
    { "UI.SCREEN",             iuiScreenNumber,      UIProperty,       thInt },
    /* @q UI.X:Int (Global Property), UI.Y:Int (Global Property)
       Scanner position.
       - on the starchart (and player screen): same as {Chart.X}, {Chart.Y}.
       - on control screens: position of the scanner. Changing the values moves the scanner.
       @assignable
       @since PCC2 1.99.10, PCC 1.0.14 */
    { "UI.X",                  iuiScanX,             UIProperty,       thInt },
    { "UI.Y",                  iuiScanY,             UIProperty,       thInt },
};

static RandomNumberGenerator script_rng(std::time(0));

/**************************** IntGlobalContext ***************************/

IntGlobalContext::IntGlobalContext()
{ }

IntGlobalContext::~IntGlobalContext()
{ }

static int32_t
lookupGlobalProperty(const IntNameQuery& name)
{
    static const uint32_t nproperties = countof(global_mapping);
    int32_t result = global_property_names.lookup(name);
    if (result >= 0)
        return result + nproperties;

    return lookupName(name, global_mapping, countof(global_mapping));
}

int32_t
IntGlobalContext::lookup(const IntNameQuery& name)
{
    if (name.startsWith("GLOBAL.", 7)) {
        return lookupGlobalProperty(IntNameQuery(name, 7));
    } else {
        return lookupGlobalProperty(name);
    }
}

void
IntGlobalContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        /* User variable */
        global_values.set(index - nproperties, value);
    } else {
        /* Global property */
        switch (global_mapping[index].domain) {
         case GlobalProperty:
            setGlobalProperty(IntGlobalProperty(global_mapping[index].index), value);
            break;
         case UIProperty:
            IntUserInterfaceBinding::get().setProperty(IntUserInterfaceProperty(global_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    }
}

IntValue*
IntGlobalContext::get(int32_t index)
{
    /* Check user variables */
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        return IntValue::cloneOf(global_values[index - nproperties]);
    }

    /* Check global properties */
    switch (global_mapping[index].domain) {
     case GlobalProperty:
        return getGlobalProperty(IntGlobalProperty(global_mapping[index].index));
     case MyPlayerProperty:
        return getPlayerProperty(getPlayerId(), IntPlayerProperty(global_mapping[index].index));
     case UIProperty:
        return IntUserInterfaceBinding::get().getProperty(IntUserInterfaceProperty(global_mapping[index].index));
    }
    return 0;
}

IntContext*
IntGlobalContext::clone()
{
    return new IntGlobalContext();
}

GObject*
IntGlobalContext::getObject()
{
    return 0;
}

string_t
IntGlobalContext::toString(bool /*readable*/)
{
    return "#<global>";
}

void
IntGlobalContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Global;
    sv.value = 0;
}

void
IntGlobalContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumNames(global_property_names);
    acceptor.enumMapping(global_mapping, countof(global_mapping));
}

/************************** IntCCNotifyConfirmed *************************/

IntCCNotifyConfirmed::IntCCNotifyConfirmed()
{ }

IntCCNotifyConfirmed::~IntCCNotifyConfirmed()
{ }

// IntCallableValue:
void
IntCCNotifyConfirmed::call(IntExecutionContext& exc, IntDataSegment& args, bool want_result)
{
    /* @q CC$NotifyConfirmed():Bool (Internal)
       Part of the implementation of {Notify}.
       Returns true if there is a confirmed notification message for this process. */
    checkArgumentCount(args.size(), 0, 0);
    if (want_result) {
        IntNotificationMessage* msg = exc.getNotificationMessage();
        exc.pushNewValue(makeBoolValue(msg != 0 && msg->isConfirmed()));
    }
}

bool
IntCCNotifyConfirmed::isProcedureCall()
{
    return false;
}

int32_t
IntCCNotifyConfirmed::getDimension(int32_t)
{
    return 0;
}

IntContext*
IntCCNotifyConfirmed::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntCCNotifyConfirmed::toString(bool /*readable*/)
{
    return "CC$NotifyConfirmed";
}

void
IntCCNotifyConfirmed::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntCCNotifyConfirmed*
IntCCNotifyConfirmed::clone()
{
    return new IntCCNotifyConfirmed();
}


/** Get global property.
    \param igp Global property to query. */
IntValue*
getGlobalProperty(IntGlobalProperty igp)
{
    switch (igp) {
     case igpFileFormatLocal:
        /* @q System.Local:Str (Global Property)
           Local file format.
           Reports the file format PCC uses to store your player files:
           - <tt>"DOS"</tt> (same as planets.exe)
           - <tt>"Windows"</tt> (same as Winplan) */

        /* In PCC 1.x, this value determines the format of control and message outbox files.
           In PCC2, those are orthogonal. We can, however, comparatively easily determine the mailbox format. */
        if (haveDisplayedTurn()) {
            if (getCurrentTurn().getOutbox(getPlayerId()).getFormat() == GOutbox::fDos)
                return makeStringValue("DOS");
            else
                return makeStringValue("Windows");
        } else {
            return 0;
        }
     case igpFileFormatRemote:
        /* @q System.Remote:Str (Global Property)
           Remote file format.
           Reports the file format PCC uses for your turn files, i.e. what the "remote" host system sees:
           - <tt>"DOS"</tt> (same as planets.exe)
           - <tt>"Windows"</tt> (same as Winplan) */

        /* In PCC 1.x, this value is 0 for Dosplan, 1 for Winplan, and
           determines the TRN format. This is the config option. As of
           20110206, PCC2 has no such config option and always produces
           Winplan format. */
        return makeStringValue("Windows");
     case igpGameDirectory:
        /* @q System.GameDirectory:Str (Global Property)
           Game directory. EMPTY when no game loaded.

           The game directory is the directory containing the current game's files.

           @diff In PCC 1.x, it is possible to concatenate this property with a file name
           to access a file in the game directory.
           This does no longer work in PCC2.
           Use the {MakeFileName} function, as in
           | Open MakeFileName(System.GameDirectory, "file.txt") For Input As #1
           to access files in the game directory. */
        if (game_dir_name.size())
            return makeStringValue(game_dir_name);
        else
            return 0;
     case igpMyInMsgs:
        /* @q My.InMsgs:Int (Global Property)
           Number of incoming (received) messages this turn. */
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getInbox().getCount());
        else
            return 0;
     case igpMyOutMsgs:
        /* @q My.OutMsgs:Int (Global Property)
           Number of outgoing (sent) messages this turn. */
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getOutbox(getPlayerId()).getCount());
        else
            return 0;
     case igpMyVCRs:
        /* @q My.VCRs:Int (Global Property)
           Number of incoming combat recordings this turn. */
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getVcrs().getNumBattles());
        else
            return 0;
     case igpRootDirectory:
        /* @q System.RootDirectory:Str (Global Property)
           Root directory.

           The root directory is the directory within the program installation directory
           containing the default specification files.
           If a specification file cannot be found in the {System.GameDirectory|game directory},
           it is looked for in the root directory.
           This directory typically is one of
           - /usr/local/share/planets
           - C:\Programs\PCC2\specs

           @diff In PCC 1.x, it is possible to concatenate this property with a file name
           to access a file in the root directory.
           This does no longer work in PCC2.
           Use the {MakeFileName} function, as in
           | Open MakeFileName(System.RootDirectory, "file.txt") For Input As #1
           to access files in the root directory. */
        if (root_dir_name.size())
            return makeStringValue(root_dir_name);
        else
            return 0;
     case igpSelectionLayer:
        /* @q Selection.Layer:Int (Global Property)
           Current selection layer.
           A number from 0 to 7.
           @assignable */
        return makeIntValue(GMultiSelection::getCurrentSelectionLayer());
     case igpSystemLanguage:
        /* @q System.Language:Str (Global Property)
           Language code.
           This is the language the user wants to use,
           usually in the form of a two-letter ISO 639 code ("en" = English).
           @since PCC2 1.99.25 */
        return makeStringValue(getLanguageCode());
     case igpSystemProgram:
        /* @q System.Program:Str (Global Property)
           Name of the program executing the script.
           Values in use so far:
           - <tt>"PCC"</tt> (PCC 1.x or PCC2 desktop application)
           - <tt>"PCC (Standalone Interpreter)"</tt> (PCC 1.x standalone interpreter) */
        return makeStringValue("PCC");
     case igpSystemVersion:
        /* @q System.Version:Str (Global Property)
           Version number of the program executing the script.
           For example, "1.1.18", or "1.99.20". */
        return makeStringValue(PCC2_VERSION);
     case igpSystemVersionCode:
        /* @q System.Version$:Int (Global Property)
           Version number of the program executing the script.
           The version number has three digits for the "patchlevel" field, and two digits for the "minor" field.
           For example, "101018" or "199020" for "1.1.18" and "1.99.20", respectively. */
        return makeIntValue(PCC2_VERSION_CODE);
     case igpSystemHost:
        /* @q System.Host:Str (Global Property)
           Name of Host program.
           Values in use so far:
           - <tt>"PHost"</tt>
           - <tt>"SRace"</tt>
           - <tt>"Host"</tt> */
        if (host.isPHost())
            return makeStringValue("PHost");
        else if (host.isSRace())
            return makeStringValue("SRace");
        else if (host.isTHost())
            return makeStringValue("Host");
        else
            return 0;
     case igpSystemHostCode:
        /* @q System.Host$:Int (Global Property)
           Name of Host program.
           Values in use so far:
           <table>
            <tr><th width="7" align="left">System.Host</th><th width="8" align="left">System.Host$</th></tr>
            <tr><td width="7">Host</td>       <td width="8">0</td></tr>
            <tr><td width="7">SRace</td>      <td width="8">1</td></tr>
            <tr><td width="7">PHost</td>      <td width="8">2</td></tr>
           </table>
           You should prefer using {System.Host} instead. */
        if (host.isPHost())
            return makeIntValue(2);
        else if (host.isSRace())
            return makeIntValue(1);
        else if (host.isTHost())
            return makeIntValue(0);
        else
            return 0;
     case igpSystemHostVersion:
        /* @q System.HostVersion:Int (Global Property)
           Host version number.
           The version number is converted to a number, with three digits fo the patchlevel and two for the minor version.
           If the respective host version uses letters to specify the patchlevel, "a" is 1, "b" is 2, and so on.
           PCC2 also knows that some host versions use "3.1" to actually mean "3.10".
           Examples:
           <table>
            <tr><th width="5">Version</th><th width="5">Value</th></tr>
            <tr><td width="5">3.22.20</td><td width="5">322020</td></tr>
            <tr><td width="5">3.15</td>   <td width="5">315000</td></tr>
            <tr><td width="5">3.5c</td>   <td width="5">305003</td></tr>
           </table> */
        return makeIntValue(host.host_version);
     case igpRegSharewareFlag:
        /* @q System.GameType$:Bool (Global Property)
           Registration flag.
           %True if you use a shareware key (Tech 6 limit), %False if you use a full version. */
        return makeBoolValue(!getRegInfo().isRegistered());
     case igpRegSharewareText:
        /* @q System.GameType:Str (Global Property)
           Registration flag.
           One of <tt>"Shareware"</tt> or <tt>"Registered"</tt>. */
        return makeStringValue(getRegInfo().isRegistered() ? "Registered" : "Shareware");
     case igpRandomSeed:
        /* @q System.RandomSeed:Int (Global Property)
           Random number generator seed.
           Using the same seed, you can reproduce the same random number sequence.
           The seed is a full 32-bit value.

           The underlying random number generator is undocumented as of now,
           and has nothing to do with the random number generators used in VCR/PVCR.
           It is not guaranteed that the same random number generator will be used throughout all versions of PCC.
           @see Random
           @assignable */
        return makeIntValue(script_rng.getSeed());
     case igpRegStr1:
        /* @q System.RegStr1:Str (Global Property)
           Your registration key.
           This is the first line (name or registration number) of the key. */
        return makeStringValue(getRegInfo().getRegStr1());
     case igpRegStr2:
        /* @q System.RegStr2:Str (Global Property)
           Your registration key.
           This is the second line (registration number or date) of the key. */
        return makeStringValue(getRegInfo().getRegStr2());
     case igpTurnNumber:
        /* @q Turn:Int (Global Property)
           Turn number. */
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getTurnNumber());
        else
            return 0;
     case igpTurnIsNew:
        /* @q Turn.IsNew:Bool (Global Property)
           New-turn flag.
           True if this is a new turn, false if you have opened PCC for the second time this turn. */
        if (haveDisplayedTurn())
            return makeBoolValue(getDisplayedTurn().getDatabaseTurn() < getDisplayedTurn().getTurnNumber());
        else
            return 0;
     case igpTurnTime:
        /* @q Turn.Time:Str (Global Property)
           Turn time.
           Time of last host run, in <tt>hh:mm:ss</tt> format,
           using the host's timezone and 24-hour format. */
        if (haveDisplayedTurn())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getTimestamp().getTimeAsString());
        else
            return 0;
     case igpTurnDate:
        /* @q Turn.Date:Str (Global Property)
           Turn date.
           Date of last host run, in <tt>mm-dd-yyyy</tt> format,
           using the host's timezone. */
        if (haveDisplayedTurn())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getTimestamp().getDateAsString());
        else
            return 0;
    }
    return 0;
}

/** Set global property.
    \param igp Global property to set
    \param value New value */
void
setGlobalProperty(IntGlobalProperty igp, IntValue* value)
{
    int32_t iv;
    switch (igp) {
     case igpSelectionLayer:
        if (checkIntArg(iv, value, 0, GMultiSelection::NUM_SELECTION_LAYERS-1))
            GMultiSelection::setCurrentSelectionLayer(iv);
        break;
     case igpRandomSeed:
        if (checkIntArg(iv, value))
            script_rng.setSeed(iv);
        break;
     default:
        throw IntError::notAssignable();
    }
}

/******************************* AuthPlayer ******************************/

static GPlayerSet saved_password_players;
static string_t saved_passwords[NUM_PLAYERS];

bool
checkSavedPassword(int32_t player, const GGen& gen)
{
    return player > 0
        && player <= NUM_PLAYERS
        && saved_password_players.contains(player)
        && gen.isPassword(saved_passwords[player-1]);
}

void
clearSavedPasswords()
{
    saved_password_players.clear();
}

/******************************* Functions *******************************/

/* @q Distance(obj1:Any, x2:Int, y2:Int):Num (Function)
   @noproto
   | Distance(x1:Int, y1:Int, x2:Int, y2:Int):Num
   | Distance(x1:Int, y1:Int, obj2:Any):Num
   | Distance(obj1:Any, x2:Int, y2:Int):Num
   | Distance(obj1:Any, obj2:Any):Num
   Compute distance between two points.
   Points can be specified as two integers for an X/Y coordinate pair,
   or an object which must have {Loc.X} and {Loc.Y} properties.
   Examples:
   | Distance(1000, 1000, 1200, 1200)
   | Distance(Ship(10), Planet(30))

   If a wrapped map is being used, the map seam is also considered and the shortest possible distance is reported.
   @since PCC 1.0.11, PCC2 1.99.8 */
IntValue*
IFDistanceGet(IntArgBlock& args)
{
    GPoint points[2];

    for (int i = 0; i < 2; ++i) {
        /* Do we have an argument? */
        if (args.getNumArgs() == 0)
            throw IntError("Too few arguments for \"Distance\"");
        IntValue* theValue = args.getNext();

        /* What is it? */
        if (theValue == 0) {
            /* Null. Result is null. */
            return 0;
        } else if (IntContext* cv = dynamic_cast<IntContext*>(theValue)) {
            /* Context. Must have LOC.X and LOC.Y properties. */
            int32_t x = cv->lookup(IntNameQuery("LOC.X"));
            int32_t y = cv->lookup(IntNameQuery("LOC.Y"));
            if (x < 0 || y < 0)
                throw IntError("Operand doesn't have a position");

            /* Fetch values */
            std::auto_ptr<IntValue> vx(cv->get(x));
            std::auto_ptr<IntValue> vy(cv->get(y));
            int32_t ix, iy;
            if (!checkIntArg(ix, vx.get()) || !checkIntArg(iy, vy.get()))
                return 0;

            points[i].x = ix;
            points[i].y = iy;
        } else {
            /* Possibly integer. There must be another integer. */
            if (args.getNumArgs() == 0)
                throw IntError("Too few arguments for \"Distance\"");
            IntValue* otherValue = args.getNext();

            /* Null? */
            int32_t ix, iy;
            if (!checkIntArg(ix, theValue) || !checkIntArg(iy, otherValue))
                return 0;

            points[i].x = ix;
            points[i].y = iy;
        }
    }

    if (args.getNumArgs() != 0)
        throw IntError("Too many arguments to \"Distance\"");

    return makeFloatValue(std::sqrt(double(distanceSquared(points[0], points[1]))));
}

/* @q ShipName(sid:Int):Str (Function)
   Get name of a ship.
   Similar to <tt>Ship(sid).Name</tt>.

   @diff In PCC 1.x, this function returns a decorated name such as "USS Lincoln (#123)".
   In PCC2 since 1.99.20, this function returns just the name.
   Use {ShipNameAndId} for the old behaviour.
   @see ShipNameAndId, PlanetName
   @since PCC 0.99.7, PCC2 1.99.8 */
IntValue*
IFShipNameGet(IntArgBlock& args)
{
    int32_t sid;

    args.checkArgumentCount(1);
    if (!checkIntArg(sid, args.getNext()))
        return 0;

    if (sid <= 0 || sid > NUM_SHIPS)
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sid).getName(GObject::PlainName));
}

/* @q PlanetName(pid:Int):Str (Function)
   Get name of a planet.
   @see ShipName
   @since PCC 0.99.7, PCC2 1.99.8  */
IntValue*
IFPlanetNameGet(IntArgBlock& args)
{
    int32_t sid;

    args.checkArgumentCount(1);
    if (!checkIntArg(sid, args.getNext()))
        return 0;

    if (sid <= 0 || sid > NUM_PLANETS)
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sid).getName(GObject::PlainName));
}

/* @q PlanetAt(x:Int, y:Int, Optional flag:Bool):Int (Function)
   Get planet by location.

   Returns the Id number of the planet at position (%x,%y).
   When %flag is True (nonzero, nonempty), returns the planet whose gravity wells are in effect at that place;
   when %flag is False or not specified at all, returns only exact matches.
   If there is no such planet, it returns zero.
   @since PCC 1.0.18, PCC2 1.99.9 */
IntValue*
IFPlanetAtGet(IntArgBlock& args)
{
    int32_t x, y;

    // Fetch x,y parameters
    args.checkArgumentCount(2, 3);
    if (!checkIntArg(x, args.getNext(), -32767, 32767) || !checkIntArg(y, args.getNext(), -32767, 32767))
        return 0;

    // Fetch optional flag argument
    bool flag = false;
    if (args.getNumArgs() > 0) {
        int n = getBoolValue(args.getNext());
        if (n < 0)
            return 0;
        flag = (n != 0);
    }

    // Generate result
    if (!haveDisplayedTurn())
        return 0;

    return makeIntValue(getDisplayedTurn().getCurrentUniverse().getPlanetAt(GPoint(x, y), flag));
}

static bool fetchPlayerArg(int& player, IntArgBlock& args)
{
    if (args.getNumArgs() > 0) {
        return checkIntArg(player, args.getNext(), 1, NUM_PLAYERS);
    } else {
        player = getPlayerId();
        return true;
    }
}


static IntValue* getConfigValue(ConfigOption* opt, IntArgBlock& args)
{
    /* Figure out option */
    if (ConfigAlias* a = dynamic_cast<ConfigAlias*>(opt))
        opt = &a->getForwardedOption();

    if (ConfigIntArrayBaseOption* bopt = dynamic_cast<ConfigIntArrayBaseOption*>(opt)) {
        /* ints, optional player */
        /* Possible limits are
           2    NewNativesPopulationRange
           4    WraparoundRectangle
           8    MeteorShowerOreRanges
           9    NewNativesRaceRate
           10   ConfigExpOption, e.g. EModBayRechargeRate
           11   ConfigStdOption, e.g. RaceMiningRate */
        int player;
        if (bopt->getLimit() == 11) {
            if (!fetchPlayerArg(player, args)) {
                return 0;
            }
        } else {
            args.checkArgumentCount(1);
            if (!checkIntArg(player, args.getNext(), 1, bopt->getLimit())) {
                return 0;
            }
        }
        return makeIntValue(bopt->getArray()[player-1]);
    } else if (ConfigIntOption* intopt = dynamic_cast<ConfigIntOption*>(opt)) {
        /* single int, no player. Example: NumShips */
        args.checkArgumentCount(0);
        return makeIntValue((*intopt)());
    } else if (ConfigIntRefOption* iropt = dynamic_cast<ConfigIntRefOption*>(opt)) {
        /* integer reference, e.g. Chart.Small.Show */
        args.checkArgumentCount(0);
        return makeIntValue((*iropt)());
    } else if (ConfigCostOption* costopt = dynamic_cast<ConfigCostOption*>(opt)) {
        /* 11 costs, optional player. Example: StarbaseCost */
        int player;
        if (!fetchPlayerArg(player, args)) {
            return 0;
        }
        return makeStringValue((*costopt)(player).toCargospec());
    } else if (ConfigStringOption* stropt = dynamic_cast<ConfigStringOption*>(opt)) {
        /* single string, no player */
        args.checkArgumentCount(0);
        return makeStringValue((*stropt)());
    } else if (ConfigStringArrayOption* saopt = dynamic_cast<ConfigStringArrayOption*>(opt)) {
        /* string array, applies to Language, ExperienceLevelNames. Parameter must be given. */
        if (args.getNumArgs() > 0) {
            int player;
            args.checkArgumentCount(1);
            if (!checkIntArg(player, args.getNext(), saopt->getFirstIndex(), saopt->getFirstIndex() + saopt->getNumElements() - 1)) {
                return 0;
            }
            return makeStringValue((*saopt)(player));
        } else {
            return makeStringValue((*saopt).toString());
        }
    } else {
        /* unknown, for example, ConfigMarkerOption */
        return makeStringValue(opt->toString());
    }
}

/* @q Cfg(key:Str, Optional player:Int):Any (Function)
   Access host configuration.
   The first parameter is the name of a configuration option as used in <tt>pconfig.src</tt>,
   such as "AllowHiss" or "UnitsPerTorpRate".
   The function returns the value of this option, an integer or string.

   If the option is an array option, the second parameter can be specified to determine which player's value to get.
   When the second parameter is not specified for an array option, the return value is the value for your race.

   @diff This function was available with a different, more complicated definition in PCC 0.98.5 up to 1.0.8,
   under the names %Cfg and %CfgL.

   @diff The detail behaviour for array options differs between versions. Since 2.0.7, the behaviour is as follows:
   - for regular per-player options, the second argument is optional and defaults to the current player.
     If EMPTY is given, the result is empty. Earlier versions of PCC2 reject EMPTY.
   - for ExperienceLevelNames and Language, if the second argument is given, the respective element is returned.
     If no second argument is given, the entire option is returned as a string.
     Versions before 2.0.7 only support the no-argument version.
   - for other array options, the argument is mandatory; passing EMPTY yields empty.
   PCC 1.x behaves subtly different, which is often considered a bug.

   @since PCC 1.0.9, PCC2 1.99.8 */
IntValue*
IFCfgGet(IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    /* Config key */
    string_t optName;
    if (!checkStringArg(optName, args.getNext()))
        return 0;
    ConfigOption* opt = config.getConfigOption(optName);
    if (opt == 0)
        throw IntError("Invalid first argument to \"Cfg\"");

    return getConfigValue(opt, args);
}

/* @q Pref(key:Str, Optional index:Int):Any (Function)
   Access user configuration (preferences).
   The first parameter is the name of a preference setting, such as "Backup.Turn" or "Label.Ship".
   The function returns the value of this option, an integer, boolean or string.

   If the option is an array option, the second parameter must be specified as the index into the array,
   starting at 1.

   @since PCC2 2.0.12 */
IntValue*
IFPrefGet(IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    /* Config key */
    string_t optName;
    if (!checkStringArg(optName, args.getNext()))
        return 0;
    ConfigOption* opt = getUserPreferences().getConfigOption(optName);
    if (opt == 0)
        throw IntError("Invalid first argument to \"Pref\"");

    return getConfigValue(opt, args);
}

/* @q IsSpecialFCode(fc:Str):Bool (Function)
   Check for special friendly code.
   Returns true if the friendly code given as a parameter is a special friendly code.

   A special friendly code is one defined as special through the <tt>fcodes.cc</tt> file, or through <tt>xtrfcode.txt</tt>.
   Note that PCC2 before 2.0.8 does not consider <tt>xtrfcode.txt</tt>.

   @since PCC 1.1.4, PCC2 1.99.8 */
IntValue*
IFIsSpecialFCodeGet(IntArgBlock& args)
{
    /* Check arguments */
    args.checkArgumentCount(1);
    IntValue* arg = args.getNext();
    if (!arg)
        return 0;

    /* Function implementation */
    string_t fc = arg->toString(false).substr(0, 3);
    return makeBoolValue(GFCode::isSpecial(fc, true) || GFCode::isExtraFC(fc));
}

/* @q Truehull(slot:Int, Optional player:Int):Int (Function)
   Access per-player hull assignments.
   Returns the Id of the %slot'th hull number the specified %player can build.
   If the %player parameter is omitted, uses your player slot.
   If the specified slot does not contain a buildable hull, returns 0.
   @since PCC 1.0.12, PCC2 1.99.8 */
IntValue*
IFTruehull(IntArgBlock& args)
{
    int32_t slot, player;

    /* Check arguments */
    args.checkArgumentCount(1, 2);
    if (!checkIntArg(slot, args.getNext()))
        return 0;

    if (args.getNumArgs() > 0) {
        if (!checkIntArg(player, args.getNext()))
            return 0;
    } else {
        player = getPlayerId();
    }

    int result;
    if (slot <= 0 || slot > NUM_HULLS_PER_PLAYER || player <= 0 || player > NUM_PLAYERS)
        result = 0;
    else
        result = getTruehull(player, slot);

    return makeIntValue(result);
}

/* @q RandomFCode():Str (Function)
   Generate a random friendly code.
   The friendly code will not have a special meaning.
   @since PCC 1.1.11, PCC2 1.99.8 */
IntValue*
IFRandomFCode(IntArgBlock& args)
{
    args.checkArgumentCount(0);
    return makeStringValue(GFCode::generateRandomFCode());
}

/* @q Format(fmt:Str, args:Any...):Str (Function)
   Format a string.
   The format string can contain placeholders, each of which is replaced by one of the arguments,
   similar to the %sprintf function found in many programming languages.

   Some placeholders:
   - <tt>&#37;d</tt> formats an integer as a decimal number ("99")
   - <tt>&#37;e</tt> formats a fraction in exponential format ("9.99e+1")
   - <tt>&#37;f</tt> formats a fraction as regular decimal fraction ("99.9")
   - <tt>&#37;g</tt> auto-selects between <tt>&#37;e</tt> and <tt>&#37;f</tt>
   - <tt>&#37;x</tt> formats an integer as an octal number ("143")
   - <tt>&#37;s</tt> formats a string
   - <tt>&#37;x</tt> formats an integer as a hexadecimal number ("63")

   You can specify a decimal number between the percent sign and the letter
   to format the result with at least that many places.

   This function supports up to 10 arguments (plus the format string) in one call.

   @since PCC2 1.99.9
   @see RXml */
IntValue*
IFFormat(IntArgBlock& args)
{
    enum { LIMIT = 10 };
    args.checkArgumentCount(1, LIMIT+1);

    // First, find the format string
    string_t fmt;
    if (!checkStringArg(fmt, args.getNext()))
        return 0;

    // The formatter is intended to be used as a temporary, and keeps references
    // to its arguments. So we must make sure they live sufficiently long.
    int32_t ints[LIMIT];
    double floats[LIMIT];
    string_t strings[LIMIT];
    format formatter(fmt.c_str());

    for (uint32_t i = 0, limit = args.getNumArgs(); i < limit; ++i) {
        // Check and convert value
        IntValue* v = args.getNext();
        if (v == 0)
            return 0;

        if (IntIntValue* iv = dynamic_cast<IntIntValue*>(v)) {
            ints[i] = iv->getValue();
            formatter << ints[i];
        } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(v)) {
            floats[i] = fv->getValue();
            formatter << floats[i];
        } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(v)) {
            strings[i] = sv->getValue();
            formatter << strings[i];
        } else {
            throw IntError::typeError(IntError::tyBaseType);
        }
    }

    // Format
    return makeStringValue(formatter);
}

/* @q Translate(str:Str):Str (Function)
   Translate a string.
   Uses PCC's internal language database to reproduce the English string given as parameter
   in the user's preferred language.
   If the string is not contained in the language database, returns the original string.
   @since PCC2 1.99.9 */
IntValue*
IFTranslate(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;
    return makeStringValue(_(s.c_str()));
}

/* @q Random(a:Int, Optional b:Int):Int (Function)
   Generate random number.
   With one parameter, generates a random number in range [0,a) (i.e. including zero, not including %a).
   With two parameters, generates a random number in range [a,b) (i.e. including %a, not including %b).

   For example, <tt>Random(10)</tt> generates random numbers between 0 and 9, as does <tt>Random(0, 10)</tt>.

   <tt>Random(1,500)</tt> generates random numbers between 1 and 499,
   <tt>Random(500,1)</tt> generates random numbers between 2 and 500
   (the first parameter always included in the range, the second one is not).

   @since PCC 1.0.7, PCC2 1.99.9 */
IntValue*
IFRandom(IntArgBlock& args)
{
    int32_t lo, hi;
    args.checkArgumentCount(1, 2);
    if (!checkIntArg(lo, args.getNext(), 0, 0x7FFF))
        return 0;

    if (args.getNumArgs() > 0) {
        if (!checkIntArg(hi, args.getNext(), 0, 0x7FFF))
            return 0;
    } else {
        hi = lo;
        lo = 0;
    }

    if (lo < hi) {
        return makeIntValue(lo + script_rng(hi - lo));
    } else if (lo > hi) {
        return makeIntValue(lo - script_rng(lo - hi));
    } else {
        return makeIntValue(lo);
    }
}

/* @q CAdd(a:Cargo...):Cargo (Function)
   Add cargo sets.
   Returns a new cargo set containing the sum of all cargo sets given as parameter.
   @diff PCC 1.x supports two to six arguments for this function, PCC2 supports any number from one up.
   @since PCC 1.0.10, PCC2 1.99.9
   @see CSub */
IntValue*
IFCAdd(IntArgBlock& args)
{
    // At least one arg (PCC 1.x: at least two)
    args.checkArgumentCountAtLeast(1);

    GCargoSpec sum;
    while (args.getNumArgs() != 0) {
        GCargoSpec summand;
        if (!checkCargoSpecArg(summand, args.getNext()))
            return 0;
        sum += summand;
    }

    return makeStringValue(sum.toCargoSpecString());
}

/* @q CCompare(a:Cargo, b:Cargo):Bool (Function)
   Compare cargo sets.
   Returns true if %a contains enough cargo to remove %b.
   Supply sale is taken into account.
   @since PCC 1.0.10, PCC2 1.99.9
   @see CSub */
IntValue*
IFCCompare(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    GCargoSpec a, b;
    if (!checkCargoSpecArg(a, args.getNext()))
        return 0;
    if (!checkCargoSpecArg(b, args.getNext()))
        return 0;

    return makeBoolValue(a.isEnoughFor(b));
}

/* @q CDiv(a:Cargo, n:Int):Cargo (Function)
   @noproto
   | CDiv(a:Cargo, n:Int):Cargo
   | CDiv(a:Cargo, b:Cargo):Int
   Divide cargo sets.

   In the first form, tries to divide the cargo set %a into %n equal parts, and returns the size of one part,
   as a cargo set.

   In the second form, tries to determine how many times %b can be removed from %a.
   Supply sale is taken into account.

   @since PCC 1.1.17, PCC2 1.99.9 */
IntValue*
IFCDiv(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    GCargoSpec a;
    if (!checkCargoSpecArg(a, args.getNext()))
        return 0;

    IntValue* arg = args.getNext();
    if (arg == 0) {
        // Null
        return 0;
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg)) {
        // Divide cargospec by integer
        if (!a.divide(iv->getValue()))
            throw IntError("Divide by zero");
        return makeStringValue(a.toCargoSpecString());
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg)) {
        // Divide cargospec by cargospec
        GCargoSpec b(sv->getValue(), false);
        int32_t result;
        if (!a.divide(b, result))
            throw IntError("Divide by zero");
        return makeIntValue(result);
    } else {
        throw IntError::typeError(IntError::tyBaseType);
    }
}

/** Common back-end for CExtract and CRemove.
    \param cs [in/out] CargoSpec. On output, the result of CRemove(cs, s).
    \param total [out] On output, result of CExtract(cs, s).
    \param s [in] Type specification */
static void
doCExtractRemove(GCargoSpec& cs, int32_t& total, const string_t& s)
{
    total = 0;
    for (string_t::size_type i = 0; i < s.size(); ++i) {
        if (uint8_t(s[i]) > ' ') {
            int ty = GCargoSpec::charToCargoType(s[i]);
            if (ty < 0)
                throw IntError("Invalid cargo type");
            total += cs.get(GCargoType(ty));
            cs.set(GCargoType(ty), 0);
        }
    }
}

/* @q CExtract(a:Cargo, ele:Str):Int (Function)
   Extract cargo.
   %a is a cargo set, %ele is a cargo type (e.g. <tt>"n"</tt> for Neutronium).
   This function returns the amount of that cargo type in the cargo set.
   If multiple cargo types are given, their amounts are summed up, e.g.
   | CExtract(e, "s$")
   will report the total number of supplies and money in cargo set %e.
   @since PCC 1.0.10, PCC2 1.99.9 */
IntValue*
IFCExtract(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;

    // Compute result
    int32_t result;
    doCExtractRemove(cs, result, s);

    return makeIntValue(result);
}

/* @q CMul(a:Cargo, n:Int):Cargo (Function)
   Multiply cargo set.
   Returns a new cargo set containing %n times %a.
   @since PCC 1.0.10, PCC2 1.99.9 */
IntValue*
IFCMul(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    int32_t n;
    if (!checkIntArg(n, args.getNext()))
        return 0;

    cs *= n;
    return makeStringValue(cs.toCargoSpecString());
}

/* @q CRemove(a:Cargo, ele:Str):Cargo (Function)
   Remove cargo.
   %a is a cargo set, %ele is a cargo type (e.g. <tt>"n"</tt> for Neutronium).
   This function returns a cargo set with all cargo of the specified type removed.
   %ele can also contain multiple cargo types to remove.
   @since PCC 1.0.10, PCC2 1.99.9 */
IntValue*
IFCRemove(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;

    // Compute result
    int32_t tmp;
    doCExtractRemove(cs, tmp, s);

    return makeStringValue(cs.toCargoSpecString());
}

/* @q CSub(a:Cargo, b:Cargo...):Cargo (Function)
   Subtract cargo sets.
   Removes %b and all following sets from %a, and returns the result.
   Supply sale is taken into account.
   @since PCC 1.0.10, PCC2 1.99.9
   @see CCompare, CAdd */
IntValue*
IFCSub(IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(2);

    GCargoSpec dif;
    if (!checkCargoSpecArg(dif, args.getNext()))
        return 0;

    while (args.getNumArgs() > 0) {
        GCargoSpec subtr;
        if (!checkCargoSpecArg(subtr, args.getNext()))
            return 0;
        dif -= subtr;
    }

    dif.doSupplySale();

    return makeStringValue(dif.toCargoSpecString());
}

/* @q AddCommand cmd:Str (Global Command)
   Add a command message.
   These commands are mostly for the PHost command processor.
   PCC knows how commands amend or replace each other, e.g.
   | AddCommand "allies add 3"
   will replace a previous <tt>"allies drop 3"</tt> command.
   @todo document the commands
   @since PCC 1.1.4, PCC2 1.99.9 */
void
IFAddCommand(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);

    // Fetch argument
    string_t cmdStr;
    if (!checkStringArg(cmdStr, args.getNext()))
        return;

    // Do it
    if (!haveDisplayedTurn() || !getPlayerId())
        throw IntError(GError::eUser);

    // Parse command
    GCommand* cmd = GCommand::parseCommand(cmdStr, true);
    if (!cmd)
        throw IntError("Invalid command");

    getDisplayedTurn().getCommands(getPlayerId()).setCommand(cmd);
}

/* @q AddConfig line:Str (Global Command)
   Modify the game configuration (PConfig/HConfig).
   %line is a configuration assignment as it could appear in <tt>pconfig.src</tt>.
   This command will process the line, and update the in-memory configuration accordingly
   (it will not update the configuration file!).

   For example,
   | AddConfig "EngineShieldBonusRate = 0"
   will disable the engine-shield bonus. After that command,
   | {Cfg()|Cfg}("EngineShieldBonusRate")
   will return 0.

   You can only modify complete options, there's no way to modify just one slot of an array option.

   With PHost, some host settings can be permanently modified by players by sending
   a command message (for example, the language).
   Use {AddCommand} to send these messages.

   <b>This function is for people who know what they're doing.</b>
   Changing the configuration will not immediately update the screen.
   Some settings known to cause trouble, in increasing order of severity:
   - %ColonistTaxRate and friends: you must call <tt>{UI.Update} 1</tt>
     to update the predictions on control screens;
   - %CPEnableRemote, %CPEnableGive: you may have to exit and re-enter the
     control screen to add/remove the respective buttons;
   - %PlayerRace, %StarbaseCost, etc.: don't even think about modifying that in mid-game.

   @since PCC 1.1.4, PCC2 1.99.25 */
void
IFAddConfig(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t text;
    if (!checkStringArg(text, args.getNext())) {
        return;
    }

    // Must have a turn
    mustHaveTurn();

    // Parse
    string_t::size_type n = text.find('=');
    if (n == string_t::npos) {
        throw IntError("Invalid configuration setting");
    }
    ConfigOption* opt = config.getConfigOption(strTrim(text.substr(0, n)));
    if (!opt) {
        throw IntError("Invalid configuration setting");
    }
    opt->set(strTrim(text.substr(n+1)));
    opt->setSource(ConfigOption::srcUser);
    config.assignDependantOptions();
}

/* @q AddFCode line:Str (Global Command)
   Add a friendly code to the <a href="pcc2:fcode">selection list</a>.
   %line is a text line as it could appear in <tt>fcodes.cc</tt>.

   For example,
   | AddFCode "cln,s-57,Clone this ship"
   will define the "cln" friendly code (this definition already appears in the default <tt>fcodes.cc</tt> by default).

   @diff In PCC 1.x, this command always adds the new code at the end.
   In PCC2, the friendly code list is always sorted alphabetically.

   @since PCC 1.1.4, PCC2 1.99.25 */
void
IFAddFCode(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t text;
    if (!checkStringArg(text, args.getNext())) {
        return;
    }

    // Must have a turn
    mustHaveTurn();

    // Parse
    string_t::size_type n = text.find(',');
    if (n == string_t::npos) {
        throw IntError("Invalid friendly code");
    }

    // Do it
    GFCodeList::getInstance().addFCode(GFCode(strTrim(string_t(text, 0, n)), strTrim(string_t(text, n+1))));
}

/* @q AuthPlayer player:Int, password:Str (Global Command)
   Defines a player password.
   When you load the specified player's data, and the password matches, PCC2 will not ask for the password.
   It is not an error to specify the wrong password with this command.

   This command can be placed in your <a href="int:statement:startup">autoexec.q</a> file in your game directory.
   For example, when you're playing the Feds, you could put the following in the game's <tt>autoexec.q</tt> file:
   |  On BeforeLoad Do AuthPlayer 1, "kirk"     % the Fed password
   This will let you open the Fed RST without being prompted for passwords on your computer
   (but everyone else on other computers without this script will still have to know it).

   Passwords are forgotten whenever you leave the <a href="pcc2:racescreen">race screen</a>,
   so you should regenerate it in the {BeforeLoad} hook.

   @since PCC 1.1.1, PCC2 1.99.25 */
void
IFAuthPlayer(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t player;
    string_t pass;
    if (!checkIntArg(player, args.getNext(), 1, NUM_PLAYERS) || !checkStringArg(pass, args.getNext())) {
        return;
    }

    // Remember password
    saved_passwords[player-1] = pass;
    saved_password_players += player;
}

/* @q NewCircle x:Int, y:Int, radius:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new circle drawing.
   The circle will be centered at %x,%y, and have the specified %radius.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewLine, NewRectangle, NewMarker
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFNewCircle(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // NewCircle x,y,radius,[color,tag,expire]
    args.checkArgumentCount(3, 6);

    int32_t x, y, radius;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000)
        || !checkIntArg(radius, args.getNext(), 1, 5000))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Context check
    if (!haveDisplayedTurn() || !getPlayerId())
        throw IntError(GError::eUser);

    // Do it
    GDrawing drawing(GPoint(x, y), CircleDrawing);
    drawing.setCircleRadius(radius);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

static void
doLineRectangle(IntArgBlock& args, GDrawingType drawingType, bool normalizeCoords)
{
    // <command> x1,y1,x2,y2[,color,tag,expire]
    args.checkArgumentCount(4, 7);

    int32_t x1, y1, x2, y2;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;

    if (!checkIntArg(x1, args.getNext(), 0, 10000)
        || !checkIntArg(y1, args.getNext(), 0, 10000)
        || !checkIntArg(x2, args.getNext(), 0, 10000)
        || !checkIntArg(y2, args.getNext(), 0, 10000))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Refuse making drawing of a certain size
    if (std::abs(x2-x1) > 5000 || std::abs(y2-y1) > 5000)
        throw IntError::rangeError();

    // Context check
    if (!haveDisplayedTurn() || !getPlayerId())
        throw IntError(GError::eUser);

    // Normalize coordinates if needed
    GPoint a(x1, y1);
    GPoint b(x2, y2);
    if (normalizeCoords)
        b = b.getSimpleNearestAlias(a);

    // Draw it
    GDrawing drawing(a, drawingType);
    drawing.setPos2(b);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

/* @q NewRectangle x1:Int, y1:Int, x2:Int, y2:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new rectangle drawing.
   On a wrapped map, the coordinates will be adjusted so that the rectangle spans the minimum area,
   possibly by crossing a map seam.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewCircle, NewLine, NewRectangleRaw, NewMarker
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFNewRectangle(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, RectangleDrawing, true);
}

/* @q NewRectangleRaw x1:Int, y1:Int, x2:Int, y2:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new rectangle drawing.
   The coordinates will not be adjusted for wrapped maps.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewCircle, NewLine, NewRectangle, NewMarker
   @since PCC2 1.99.9, PCC 1.1.15 */
void
IFNewRectangleRaw(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, RectangleDrawing, false);
}

/* @q NewLine x1:Int, y1:Int, x2:Int, y2:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new line drawing.
   On a wrapped map, the coordinates will be adjusted so that the line covers the minimum distance,
   possibly by crossing a map seam.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewCircle, NewLineRaw, NewRectangle, NewMarker
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFNewLine(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, LineDrawing, true);
}

/* @q NewLineRaw x1:Int, y1:Int, x2:Int, y2:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new line drawing.
   The coordinates will not be adjusted for wrapped maps.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewCircle, NewLine, NewRectangle, NewMarker
   @since PCC2 1.99.9, PCC 1.1.15 */
void
IFNewLineRaw(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, LineDrawing, false);
}

/* @q NewMarker x:Int, y:Int, type:Int, Optional color:Int, tag:Int, expire:Int (Global Command)
   Create new marker drawing.
   The %type selects the marker shape.

   The %color is an integer between 0 and 30, and selects the color.
   The %tag is a value between 0 and 32767 you can use to identify your drawings,
   usually this value is created using {Atom}.

   %expire defines the time-of-expiry for the game as a turn number:
   if the current turn number is larger than this value, the drawing is automatically deleted.
   Thus, set %expire=0 to make drawings only visible for the current session.
   %expire=-1 is the default, drawings with this value never expire.
   @see NewCircle, NewLineRaw, NewRectangle, NewMarker
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFNewMarker(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // NewMarker x,y,typ[,color,text,tag,expire]
    args.checkArgumentCount(3, 7);

    int32_t x, y, type;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;
    string_t text;
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000)
        || !checkIntArg(type, args.getNext(), 0, NUM_USER_MARKERS-1))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkStringArg(text, args.getNext());
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Context check
    if (!haveDisplayedTurn() || !getPlayerId())
        throw IntError(GError::eUser);

    // Draw it
    GDrawing drawing(GPoint(x, y), MarkerDrawing);
    drawing.setMarkerKind(type);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);
    drawing.setComment(text);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

/* @q SaveGame (Global Command)
   Save current game. */
void
IFSaveGame(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (GGameDataStorage::instance)
        GGameDataStorage::instance->save(getCurrentTurn());
    else
        throw IntError("No game loaded");
}

namespace {
    class SelectionLoadAskUI : public AnswerProvider {
     public:
        bool ask(int, string_t q);
    };
}

bool
SelectionLoadAskUI::ask(int, string_t q)
{
    /* This is wild. We have to ask the user a question. To do that,
       we need to call UI.Message, because only that knows whether
       we're in the GUI or on the console. As an added bonus, it is
       implemented as a script on the console, so we cannot just look
       up a symbol and call that natively. So, we have to create a
       fully-fledged (temporary) process running a newly-made BCO.

          pushlit question
          pushlit title
          pushlit "Yes No"
          pushvar "UI.MESSAGE"
          procind 3
          pushloc "UI.RESULT"
          pushint 1
          bcmpeq
    */
    IntStringValue svq((format(_("%s. Do you want to load this file?"), q)));
    IntStringValue svt(_("Load Selection"));
    IntStringValue svb(_("Yes No"));
    IntBCORef bco = new IntBytecodeObject();
    bco->setIsProcedure(false);
    bco->addLocalVariable("UI.RESULT");
    bco->addPushLiteral(&svq);
    bco->addPushLiteral(&svt);
    bco->addPushLiteral(&svb);
    bco->addVariableReferenceInstruction(IntOpcode::maPush, "UI.MESSAGE", IntCompilationContext());
    bco->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 3);
    bco->addVariableReferenceInstruction(IntOpcode::maPush, "UI.RESULT", IntCompilationContext());
    bco->addInstruction(IntOpcode::maPush, IntOpcode::sInteger, 1);
    bco->addInstruction(IntOpcode::maBinary, biCompareEQ_NC, 0);

    IntExecutionContext exc("Temporary: SelectionLoad");
    IntExecutionFrame* frame = new IntExecutionFrame(bco);
    frame->want_result = true;
    exc.pushNewFrame(frame);

    return runTemporaryProcess(exc) && getBoolValue(exc.getResult()) > 0;
}

/* @q SelectionLoad file:File, Optional flags:Str (Global Command)
   Load selection from file.

   The %flags argument is a combination of the following options:
   - %t ("timeless") to ignore the timestamp of the selection file and load the file even if it doesn't match
   - %a ("all") to accept loading files that contain all selection layers (refused by default)
   - %m ("merge") to merge the selection instead of replacing it
   - %u ("user interaction") to ask the user for confirmation before loading the file
   - a selection layer number to load the file into that layer
   @see SelectionSave, Selection.Layer
   @since PCC 1.1.3, PCC2 1.99.13 */
void
IFSelectionLoad(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // SelectionLoad "f[TAMU#]"
    //   timeless, all, merge, UI
    TextFile* tf;
    int32_t flags = 0;
    int32_t layer = GMultiSelection::getCurrentSelectionLayer();

    args.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, args.getNext()))
        return;
    checkFlagArg(flags, &layer, args.getNext(), "TAMU");
    if (layer < 0 || layer >= GMultiSelection::NUM_SELECTION_LAYERS)
        throw IntError::rangeError();

    int translatedFlags = 0;
    if (flags & 4) {
        /* M = Merge */
        translatedFlags |= selio_Merge;
    }
    if (flags & 8) {
        /* U = User interface */
        SelectionLoadAskUI a;
        loadSelection(*tf, layer, translatedFlags, a);
    } else {
        /* Not user interface, so evaluate other two flags as well */
        if (flags & 1) {
            /* T = Timeless */
            translatedFlags |= selio_Timeless;
        }
        if (flags & 2) {
            /* A = All */
            translatedFlags |= selio_AcceptAll;
        }
        translatedFlags |= selio_AcceptCurrent;
        translatedFlags |= selio_AcceptSingle;

        class Answer : public AnswerProvider {
         public:
            bool ask(int qid, string_t /*q*/) {
                if (qid & selio_Timeless) {
                    throw IntError("Stale file");
                }
                throw IntError("File doesn't match requested content");
            }
        };
        Answer a;
        loadSelection(*tf, layer, translatedFlags, a);
    }
}

/* @q SelectionSave file:File, Optional flags:Str (Global Command)
   Save selection into file.

   The %flags argument is a combination of the following options:
   - %t ("timeless") to create a file without timestamp that can be loaded in any turn
   - a selection layer number to save just that layer (default: all)
   @see SelectionLoad, Selection.Layer
   @since PCC 1.1.3, PCC2 1.99.13 */
void
IFSelectionSave(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // SelectionSave "f[T#]"
    TextFile* tf;
    int32_t flags = 0;
    int32_t layer = -1;

    args.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, args.getNext()))
        return;
    checkFlagArg(flags, &layer, args.getNext(), "T");
    if (layer < -1 || layer >= GMultiSelection::NUM_SELECTION_LAYERS)
        throw IntError::rangeError();

    saveSelection(*tf, layer, flags ? selio_Timeless : 0);
}

/* @q CC$SelectionExec layer:Int, code:Str (Internal)
   Back-end to {SelectionExec}. */
void
IFCCSelectionExec(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(2);

    int32_t layer;
    string_t code;
    if (!checkIntArg(layer, args.getNext(), 0, GMultiSelection::NUM_SELECTION_LAYERS)
        || !checkStringArg(code, args.getNext()))
        return;

    if (layer == 0)
        layer = GMultiSelection::getCurrentSelectionLayer();
    else
        --layer;

    GMultiSelection::executeSelectionExpression(layer, code);
}

/* @q CC$Notify msg:Str, associateWithProcess:Bool (Internal)
   Back-end to {Notify} and {AddNotify}. */
void
IFCCNotify(IntExecutionContext& exc, IntArgBlock& args)
{
    // Args: message, associate-with-process flag
    args.checkArgumentCount(2);

    string_t text;
    bool assoc;
    if (!checkStringArg(text, args.getNext())
        || !checkBoolArg(assoc, args.getNext()))
    {
        return;
    }

    string_t header;
    GObject* obj = exc.getInvokingObject();
    if (dynamic_cast<GPlanet*>(obj)) {
        header = format(_("(-p%04d)<<< Planet >>>\n\n"), obj->getId());
    } else if (dynamic_cast<GShip*>(obj)) {
        header = format(_("(-s%04d)<<< Ship >>>\n\n"), obj->getId());
    } else {
        header = _("(-X0000)<<< Notification >>>\n\n");
    }

    header += format(_("FROM: %s\n\n"), exc.getName());

    IntNotificationMessage* msg = IntNotificationMessageStore::instance.addNewMessage(header, text);
    if (assoc) {
        exc.setNotificationMessage(msg);
    }
}

