/**
  *  \file ui/widgets/dircrumbs.cc
  *  \brief UIDirectoryCrumbTrail widget
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/dircrumbs.h"
#include "gfx/font.h"
#include "util/nls.h"
#include "ui/uiglobal.h"

enum {
    GAP_X = 4,
    GAP_Y = 2
};

/** Constructor. */
UIDirectoryCrumbTrail::UIDirectoryCrumbTrail()
    : UIIconBox(GfxPoint(font_ems[FONT_NORMAL] * 40, font_heights[FONT_NORMAL] + GAP_Y*2))
{
    setChangeOnClick(true);
}

/** Destructor. */
UIDirectoryCrumbTrail::~UIDirectoryCrumbTrail()
{ }

int
UIDirectoryCrumbTrail::getItemWidth(int number)
{
    return fonts[FONT_NORMAL]->getTextWidth(getTextOf(number)) + GAP_X*2;
}

int
UIDirectoryCrumbTrail::getNumberOfItems()
{
    /* plus one for the "Places" entry */
    return dirs.size() + 1;
}

void
UIDirectoryCrumbTrail::drawItem(int number, GfxCanvas& can, const GfxRect& r, ItemState state)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    /* Draw the box */
    if (state == Selected) {
        drawSolidBar(ctx, r, UIColor::tc_Static);
        ctx.setColor(UIColor::tc_InvStatic);
    } else if (state == Hover) {
        drawSolidBar(ctx, r, UIColor::tc_Faded);
        ctx.setColor(UIColor::tc_Static);
    } else {
        drawBackground(ctx, r);
        ctx.setColor(UIColor::tc_Static);
    }
    
    outText(ctx, r.x + GAP_X, r.y + GAP_Y, getTextOf(number));
}

/** Set directory to display. This initializes the widget.
    \param dir Directory. Can be null to make this widget empty (i.e. display just the "Places" entry) */
void
UIDirectoryCrumbTrail::setDirectory(Ptr<Directory> dir)
{
    /* Check whether this actually is a change, to avoid recursion */
    if (dir == 0) {
        if (dirs.empty())
            return;
    } else {
        /* It's the same, if
           - the previous one was not null
           - dir is the same as the previous one
           - or dir has the same non-null name */
        if (!dirs.empty()
            && (dir == dirs.front()
                || (!dir->getName().empty()
                    && dir->getName() == dirs.front()->getName())))
            return;
    }

    dirs.clear();
    while (dir != 0) {
        dirs.push_back(dir);
        dir = dir->getParent();
    }
    onStructureChange(dirs.size());
}

/** Get item by number.
    \param number Index of referenced item
    \return Directory object corresponding to that item, or null if it's the "Places" entry */
Ptr<Directory>
UIDirectoryCrumbTrail::getItem(int number)
{
    if (number == 0)
        return 0;
    else
        return dirs[dirs.size() - number];
}

/** Get text of an entry.
    \param number Index of referenced item */
string_t
UIDirectoryCrumbTrail::getTextOf(int number)
{
    string_t result;
    if (number == 0) {
        result = _("[Places]");
    } else {
        result = Directory::getFileName(dirs[dirs.size() - number]->getName());
        if (result.size() == 0)
            result = dirs[dirs.size() - number]->getName();
    }
    return result;
}
